import abc

import aiofiles
from aiospb.messages import MetricDataTransfer

from ..data import MetricChange, DataType, PropertySet, PropertyValue
from aiofiles import open
import csv
import os.path


class HistoricalStore(abc.ABC):
    """Store historical changes to retrieve when host application is online"""

    @abc.abstractmethod
    async def save_dtos(self, changes: list[MetricDataTransfer]):
        """Save all change metrics of a scan"""

    @abc.abstractmethod
    async def load_all_dtos(self) -> list[MetricDataTransfer]:
        """Load all historical metric changes"""

    @abc.abstractmethod
    def has_dtos(self) -> bool:
        """Has data stored?"""

    @abc.abstractmethod
    async def clear(self):
        """Clear content of historical"""


class OnmemHistoricalStore(HistoricalStore):
    """Store historical changes on memory. Not recomemded if host will be stopped for a long time"""

    def __init__(self):
        self._dtos = []

    async def save_dtos(self, dtos: list[MetricChange]):
        self._dtos.extend(dtos)

    async def load_all_dtos(self) -> list[MetricDataTransfer]:
        dtos = [
            MetricDataTransfer(
                dto.timestamp,
                dto.value,
                dto.datatype,
                dto.alias,
                dto.name,
                dto.properties,
                is_historical=True,
            )
            for dto in self._dtos
        ]
        self._dtos.clear()
        return dtos

    def has_dtos(self):
        return bool(self._dtos)

    async def clear(self):
        self._dtos.clear()


class FsHistoricalStore(HistoricalStore):
    """Store historical changes in a plain file from a operating system"""

    def __init__(self, filename: str):
        self._fn = filename
        self._has_dtos = None

    async def save_dtos(self, dtos: list[MetricChange]):
        async with aiofiles.open(self._fn, "a") as f:
            lines = [self._create_line(dto) for dto in dtos]
            await f.writelines(lines)

        self._has_dtos = True

    def _create_line(self, dto: MetricDataTransfer) -> str:
        quality = dto.properties["quality"].value if "quality" in dto.properties else ""
        return f"{dto.timestamp};{dto.value};{dto.datatype.name};{quality};{dto.alias};{dto.name};{dto.is_null}"

    async def load_all_dtos(self) -> list[MetricDataTransfer]:
        dtos = []
        async with aiofiles.open(self._fn, "r") as f:
            ts, value, datatype, quality, alias, name, is_null = (
                await f.readline()
            ).split(";")

            properties = (
                PropertySet(quality=PropertyValue(int(quality), DataType.Int32))
                if quality != ""
                else PropertySet()
            )
            dtos.append(
                MetricDataTransfer(
                    int(ts),
                    DataType[datatype].convert_value(value),
                    DataType[datatype],
                    int(alias),
                    name,
                    properties,
                    is_null=(is_null == "True"),
                    is_historical=True,
                )
            )

        return dtos

    def has_dtos(self):
        if self._has_dtos is None:
            self._has_dtos = os.path.exists(self._fn)
        return self._has_dtos

    async def clear(self):
        os.remove(self._fn)
        self._has_dtos = False
